Imports System.Math

Module fft4g_h
    '-------- Real DFT / Inverse of Real DFT --------
    '    [definition]
    '        <case1> RDFT
    '            R[k] = sum_j=0^n-1 a[j]*cos(2*pi*j*k/n), 0<=k<=n/2
    '            I[k] = sum_j=0^n-1 a[j]*sin(2*pi*j*k/n), 0<k<n/2
    '        <case2> IRDFT (excluding scale)
    '            a[k] = (R[0] + R[n/2]*cos(pi*k))/2 + 
    '                   sum_j=1^n/2-1 R[j]*cos(2*pi*j*k/n) + 
    '                   sum_j=1^n/2-1 I[j]*sin(2*pi*j*k/n), 0<=k<n
    '    [usage]
    '        <case1>
    '            rdft(n, 1, a)
    '        <case2>
    '            rdft(n, -1, a)
    '    [parameters]
    '        n              :data length (int)
    '                        n >= 2, n = power of 2
    '        a[0...n-1]     :input/output data (double *)
    '                        <case1>
    '                            output data
    '                                a[2*k] = R[k], 0<=k<n/2
    '                                a[2*k+1] = I[k], 0<k<n/2
    '                                a[1] = R[n/2]
    '                        <case2>
    '                            input data
    '                                a[2*j] = R[j], 0<=j<n/2
    '                                a[2*j+1] = I[j], 0<j<n/2
    '                                a[1] = R[n/2]
    '    [remark]
    '        Inverse of 
    '            rdft(n, 1, a)
    '        is 
    '            rdft(n, -1, a)
    '            for (j = 0 j <= n - 1 j++) {
    '                a[j] *= 2.0 / n
    '            }
    '        .

    Public Const FFT_FORWARD As Integer = 1
    Public Const FFT_REVERSE As Integer = -1

    Private Const WR5000 As Double = 0.70710678118654757#
    Private Const M_PI_2 As Double = Math.PI / 2
    Private Const RDFT_LOOP_DIV As Integer = 64

    Private n As Integer

    Public Sub rdft(ByVal a() As Double, ByVal isgn As Integer)
        n = a.Length

        If isgn >= 0 Then
            If n > 4 Then
                bitrv2(a) : cftfsub(a) : rftfsub(a)
            ElseIf n = 4 Then
                cftfsub(a)
            End If

            Dim xi As Double = a(0) - a(1)
            a(0) += a(1) : a(1) = xi
        Else
            a(1) = 0.5# * (a(0) - a(1)) : a(0) -= a(1)

            If n > 4 Then
                rftbsub(a) : bitrv2(a) : cftbsub(a)
            ElseIf n = 4 Then
                cftfsub(a)
            End If
        End If
    End Sub

    Private Sub bitrv2(ByVal a() As Double)
        Dim l As Integer = n >> 2
        Dim m As Integer = 2
        Do While m < l
            l >>= 1
            m <<= 1
        Loop

        If m = l Then
            Dim j0 As Integer = 0
            For k0 As Integer = 0 To m - 1 Step 2
                Dim k As Integer = k0
                Dim i, j1, k1 As Integer
                Dim xr, xi, yr, yi As Double

                For j As Integer = j0 To j0 + k0 - 1 Step 2
                    xr = a(j) : xi = a(j + 1)
                    yr = a(k) : yi = a(k + 1)

                    a(j) = yr : a(j + 1) = yi
                    a(k) = xr : a(k + 1) = xi

                    j1 = j + m : k1 = k + 2 * m

                    xr = a(j1) : xi = a(j1 + 1)
                    yr = a(k1) : yi = a(k1 + 1)

                    a(j1) = yr : a(j1 + 1) = yi
                    a(k1) = xr : a(k1 + 1) = xi

                    j1 += m : k1 -= m

                    xr = a(j1) : xi = a(j1 + 1)
                    yr = a(k1) : yi = a(k1 + 1)

                    a(j1) = yr : a(j1 + 1) = yi
                    a(k1) = xr : a(k1 + 1) = xi

                    j1 += m : k1 += 2 * m

                    xr = a(j1) : xi = a(j1 + 1)
                    yr = a(k1) : yi = a(k1 + 1)

                    a(j1) = yr : a(j1 + 1) = yi
                    a(k1) = xr : a(k1 + 1) = xi

                    i = n >> 1
                    Do
                        k = k Xor i
                        If i <= k Then Exit Do
                        i >>= 1
                    Loop
                Next j

                j1 = j0 + k0 + m : k1 = j1 + m

                xr = a(j1) : xi = a(j1 + 1)
                yr = a(k1) : yi = a(k1 + 1)

                a(j1) = yr : a(j1 + 1) = yi
                a(k1) = xr : a(k1 + 1) = xi

                i = n >> 1
                Do
                    j0 = j0 Xor i
                    If i <= j0 Then Exit Do
                    i >>= 1
                Loop
            Next k0
        Else
            Dim j0 As Integer = 0
            For k0 As Integer = 2 To m - 1 Step 2
                Dim i As Integer = n >> 1
                Do
                    j0 = j0 Xor i
                    If i <= j0 Then Exit Do
                    i >>= 1
                Loop

                Dim k As Integer = k0
                For j As Integer = j0 To j0 + k0 - 1 Step 2
                    Dim xr As Double = a(j)
                    Dim xi As Double = a(j + 1)
                    Dim yr As Double = a(k)
                    Dim yi As Double = a(k + 1)

                    a(j) = yr : a(j + 1) = yi
                    a(k) = xr : a(k + 1) = xi

                    Dim j1 As Integer = j + m
                    Dim k1 As Integer = k + m

                    xr = a(j1) : xi = a(j1 + 1)
                    yr = a(k1) : yi = a(k1 + 1)

                    a(j1) = yr : a(j1 + 1) = yi
                    a(k1) = xr : a(k1 + 1) = xi

                    i = n >> 1
                    Do
                        k = k Xor i
                        If i <= k Then Exit Do
                        i >>= 1
                    Loop
                Next j
            Next k0
        End If
    End Sub

    Private Sub cftfsub(ByVal a() As Double)
        Dim l As Integer = 2
        If n > 8 Then
            cft1st(a)
            l = 8
            Do While ((l << 2) < n)
                cftmdl(l, a)
                l <<= 2
            Loop
        End If

        If (l << 2) = n Then
            For j As Integer = 0 To l - 1 Step 2
                Dim j1 As Integer = j + l
                Dim j2 As Integer = j1 + l
                Dim j3 As Integer = j2 + l

                Dim x0r As Double = a(j) + a(j1)
                Dim x0i As Double = a(j + 1) + a(j1 + 1)
                Dim x1r As Double = a(j) - a(j1)
                Dim x1i As Double = a(j + 1) - a(j1 + 1)
                Dim x2r As Double = a(j2) + a(j3)
                Dim x2i As Double = a(j2 + 1) + a(j3 + 1)
                Dim x3r As Double = a(j2) - a(j3)
                Dim x3i As Double = a(j2 + 1) - a(j3 + 1)

                a(j) = x0r + x2r : a(j + 1) = x0i + x2i
                a(j2) = x0r - x2r : a(j2 + 1) = x0i - x2i
                a(j1) = x1r - x3i : a(j1 + 1) = x1i + x3r
                a(j3) = x1r + x3i : a(j3 + 1) = x1i - x3r
            Next j
        Else
            For j As Integer = 0 To l - 1 Step 2
                Dim j1 As Integer = j + l

                Dim x0r As Double = a(j) - a(j1)
                Dim x0i As Double = a(j + 1) - a(j1 + 1)

                a(j) += a(j1) : a(j + 1) += a(j1 + 1)
                a(j1) = x0r : a(j1 + 1) = x0i
            Next j
        End If
    End Sub

    Private Sub cftbsub(ByVal a() As Double)
        Dim l As Integer = 2
        If n > 8 Then
            cft1st(a)
            l = 8
            Do While (l << 2) < n
                cftmdl(l, a)
                l <<= 2
            Loop
        End If

        If (l << 2) = n Then
            For j As Integer = 0 To l - 1 Step 2
                Dim j1 As Integer = j + l
                Dim j2 As Integer = j1 + l
                Dim j3 As Integer = j2 + l

                Dim x0r As Double = a(j) + a(j1)
                Dim x0i As Double = -a(j + 1) - a(j1 + 1)
                Dim x1r As Double = a(j) - a(j1)
                Dim x1i As Double = -a(j + 1) + a(j1 + 1)
                Dim x2r As Double = a(j2) + a(j3)
                Dim x2i As Double = a(j2 + 1) + a(j3 + 1)
                Dim x3r As Double = a(j2) - a(j3)
                Dim x3i As Double = a(j2 + 1) - a(j3 + 1)

                a(j) = x0r + x2r : a(j + 1) = x0i - x2i
                a(j2) = x0r - x2r : a(j2 + 1) = x0i + x2i
                a(j1) = x1r - x3i : a(j1 + 1) = x1i - x3r
                a(j3) = x1r + x3i : a(j3 + 1) = x1i + x3r
            Next j
        Else
            For j As Integer = 0 To l - 1 Step 2
                Dim j1 As Integer = j + l

                Dim x0r As Double = a(j) - a(j1)
                Dim x0i As Double = -a(j + 1) + a(j1 + 1)

                a(j) += a(j1) : a(j + 1) = -a(j + 1) - a(j1 + 1)
                a(j1) = x0r : a(j1 + 1) = x0i
            Next j
        End If
    End Sub

    Private Sub cft1st(ByVal a() As Double)
        Dim x0r As Double = a(0) + a(2)
        Dim x0i As Double = a(1) + a(3)
        Dim x1r As Double = a(0) - a(2)
        Dim x1i As Double = a(1) - a(3)
        Dim x2r As Double = a(4) + a(6)
        Dim x2i As Double = a(5) + a(7)
        Dim x3r As Double = a(4) - a(6)
        Dim x3i As Double = a(5) - a(7)

        a(0) = x0r + x2r : a(1) = x0i + x2i
        a(4) = x0r - x2r : a(5) = x0i - x2i
        a(2) = x1r - x3i : a(3) = x1i + x3r
        a(6) = x1r + x3i : a(7) = x1i - x3r

        Dim wn4r As Double = WR5000

        x0r = a(8) + a(10) : x0i = a(9) + a(11)
        x1r = a(8) - a(10) : x1i = a(9) - a(11)
        x2r = a(12) + a(14) : x2i = a(13) + a(15)
        x3r = a(12) - a(14) : x3i = a(13) - a(15)

        a(8) = x0r + x2r : a(9) = x0i + x2i
        a(12) = x2i - x0i : a(13) = x0r - x2r

        x0r = x1r - x3i : x0i = x1i + x3r

        a(10) = wn4r * (x0r - x0i) : a(11) = wn4r * (x0r + x0i)

        x0r = x3i + x1r : x0i = x3r - x1i

        a(14) = wn4r * (x0i - x0r) : a(15) = wn4r * (x0i + x0r)

        Dim ew As Double = M_PI_2 / n
        Dim kr As Integer = 0
        For j As Integer = 16 To n - 1 Step 16
            Dim kj As Integer = n >> 2
            Do
                kr = kr Xor kj
                If kj <= kr Then Exit Do
                kj >>= 1
            Loop

            Dim wk1r As Double = Cos(ew * kr)
            Dim wk1i As Double = Sin(ew * kr)
            Dim wk2r As Double = 1 - 2 * wk1i * wk1i
            Dim wk2i As Double = 2 * wk1i * wk1r
            Dim wk3r As Double = wk1r - 2 * wk2i * wk1i
            Dim wk3i As Double = 2 * wk2i * wk1r - wk1i

            x0r = a(j) + a(j + 2) : x0i = a(j + 1) + a(j + 3)
            x1r = a(j) - a(j + 2) : x1i = a(j + 1) - a(j + 3)
            x2r = a(j + 4) + a(j + 6) : x2i = a(j + 5) + a(j + 7)
            x3r = a(j + 4) - a(j + 6) : x3i = a(j + 5) - a(j + 7)

            a(j) = x0r + x2r : a(j + 1) = x0i + x2i

            x0r -= x2r : x0i -= x2i

            a(j + 4) = wk2r * x0r - wk2i * x0i : a(j + 5) = wk2r * x0i + wk2i * x0r

            x0r = x1r - x3i : x0i = x1i + x3r

            a(j + 2) = wk1r * x0r - wk1i * x0i : a(j + 3) = wk1r * x0i + wk1i * x0r

            x0r = x1r + x3i : x0i = x1i - x3r

            a(j + 6) = wk3r * x0r - wk3i * x0i : a(j + 7) = wk3r * x0i + wk3i * x0r

            x0r = wn4r * (wk1r - wk1i)

            wk1i = wn4r * (wk1r + wk1i) : wk1r = x0r
            wk3r = wk1r - 2 * wk2r * wk1i : wk3i = 2 * wk2r * wk1r - wk1i

            x0r = a(j + 8) + a(j + 10) : x0i = a(j + 9) + a(j + 11)

            x1r = a(j + 8) - a(j + 10) : x1i = a(j + 9) - a(j + 11)

            x2r = a(j + 12) + a(j + 14) : x2i = a(j + 13) + a(j + 15)

            x3r = a(j + 12) - a(j + 14) : x3i = a(j + 13) - a(j + 15)

            a(j + 8) = x0r + x2r : a(j + 9) = x0i + x2i

            x0r -= x2r : x0i -= x2i

            a(j + 12) = -wk2i * x0r - wk2r * x0i : a(j + 13) = -wk2i * x0i + wk2r * x0r

            x0r = x1r - x3i : x0i = x1i + x3r

            a(j + 10) = wk1r * x0r - wk1i * x0i : a(j + 11) = wk1r * x0i + wk1i * x0r

            x0r = x1r + x3i : x0i = x1i - x3r

            a(j + 14) = wk3r * x0r - wk3i * x0i : a(j + 15) = wk3r * x0i + wk3i * x0r
        Next j
    End Sub

    Private Sub cftmdl(ByVal l As Integer, ByVal a() As Double)
        Dim m As Integer = l << 2
        For j As Integer = 0 To l - 1 Step 2
            Dim j1 As Integer = j + l
            Dim j2 As Integer = j1 + l
            Dim j3 As Integer = j2 + l

            Dim x0r As Double = a(j) + a(j1)
            Dim x0i As Double = a(j + 1) + a(j1 + 1)
            Dim x1r As Double = a(j) - a(j1)
            Dim x1i As Double = a(j + 1) - a(j1 + 1)
            Dim x2r As Double = a(j2) + a(j3)
            Dim x2i As Double = a(j2 + 1) + a(j3 + 1)
            Dim x3r As Double = a(j2) - a(j3)
            Dim x3i As Double = a(j2 + 1) - a(j3 + 1)

            a(j) = x0r + x2r : a(j + 1) = x0i + x2i
            a(j2) = x0r - x2r : a(j2 + 1) = x0i - x2i
            a(j1) = x1r - x3i : a(j1 + 1) = x1i + x3r
            a(j3) = x1r + x3i : a(j3 + 1) = x1i - x3r
        Next j

        Dim wn4r As Double = WR5000

        For j As Integer = m To l + m - 1 Step 2
            Dim j1 As Integer = j + l
            Dim j2 As Integer = j1 + l
            Dim j3 As Integer = j2 + l

            Dim x0r As Double = a(j) + a(j1)
            Dim x0i As Double = a(j + 1) + a(j1 + 1)
            Dim x1r As Double = a(j) - a(j1)
            Dim x1i As Double = a(j + 1) - a(j1 + 1)
            Dim x2r As Double = a(j2) + a(j3)
            Dim x2i As Double = a(j2 + 1) + a(j3 + 1)
            Dim x3r As Double = a(j2) - a(j3)
            Dim x3i As Double = a(j2 + 1) - a(j3 + 1)

            a(j) = x0r + x2r : a(j + 1) = x0i + x2i
            a(j2) = x2i - x0i : a(j2 + 1) = x0r - x2r

            x0r = x1r - x3i : x0i = x1i + x3r

            a(j1) = wn4r * (x0r - x0i) : a(j1 + 1) = wn4r * (x0r + x0i)

            x0r = x3i + x1r : x0i = x3r - x1i

            a(j3) = wn4r * (x0i - x0r) : a(j3 + 1) = wn4r * (x0i + x0r)
        Next j

        Dim ew As Double = M_PI_2 / n
        Dim kr As Integer = 0
        Dim m2 As Integer = 2 * m
        For k As Integer = m2 To n - 1 Step m2
            Dim kj As Integer = n >> 2
            Do
                kr = kr Xor kj
                If kj <= kr Then Exit Do
                kj >>= 1
            Loop

            Dim wk1r As Double = Cos(ew * kr)
            Dim wk1i As Double = Sin(ew * kr)
            Dim wk2r As Double = 1 - 2 * wk1i * wk1i
            Dim wk2i As Double = 2 * wk1i * wk1r
            Dim wk3r As Double = wk1r - 2 * wk2i * wk1i
            Dim wk3i As Double = 2 * wk2i * wk1r - wk1i

            For j As Integer = k To l + k - 1 Step 2
                Dim j1 As Integer = j + l
                Dim j2 As Integer = j1 + l
                Dim j3 As Integer = j2 + l

                Dim x0r As Double = a(j) + a(j1)
                Dim x0i As Double = a(j + 1) + a(j1 + 1)
                Dim x1r As Double = a(j) - a(j1)
                Dim x1i As Double = a(j + 1) - a(j1 + 1)
                Dim x2r As Double = a(j2) + a(j3)
                Dim x2i As Double = a(j2 + 1) + a(j3 + 1)
                Dim x3r As Double = a(j2) - a(j3)
                Dim x3i As Double = a(j2 + 1) - a(j3 + 1)

                a(j) = x0r + x2r : a(j + 1) = x0i + x2i

                x0r -= x2r : x0i -= x2i

                a(j2) = wk2r * x0r - wk2i * x0i : a(j2 + 1) = wk2r * x0i + wk2i * x0r

                x0r = x1r - x3i : x0i = x1i + x3r

                a(j1) = wk1r * x0r - wk1i * x0i : a(j1 + 1) = wk1r * x0i + wk1i * x0r

                x0r = x1r + x3i : x0i = x1i - x3r

                a(j3) = wk3r * x0r - wk3i * x0i : a(j3 + 1) = wk3r * x0i + wk3i * x0r
            Next j

            Dim wr As Double = wn4r * (wk1r - wk1i)
            wk1i = wn4r * (wk1r + wk1i) : wk1r = wr
            wk3r = wk1r - 2 * wk2r * wk1i : wk3i = 2 * wk2r * wk1r - wk1i

            For j As Integer = k + m To l + (k + m) - 1 Step 2
                Dim j1 As Integer = j + l
                Dim j2 As Integer = j1 + l
                Dim j3 As Integer = j2 + l

                Dim x0r As Double = a(j) + a(j1)
                Dim x0i As Double = a(j + 1) + a(j1 + 1)
                Dim x1r As Double = a(j) - a(j1)
                Dim x1i As Double = a(j + 1) - a(j1 + 1)
                Dim x2r As Double = a(j2) + a(j3)
                Dim x2i As Double = a(j2 + 1) + a(j3 + 1)
                Dim x3r As Double = a(j2) - a(j3)
                Dim x3i As Double = a(j2 + 1) - a(j3 + 1)

                a(j) = x0r + x2r : a(j + 1) = x0i + x2i

                x0r -= x2r : x0i -= x2i

                a(j2) = -wk2i * x0r - wk2r * x0i : a(j2 + 1) = -wk2i * x0i + wk2r * x0r

                x0r = x1r - x3i : x0i = x1i + x3r

                a(j1) = wk1r * x0r - wk1i * x0i : a(j1 + 1) = wk1r * x0i + wk1i * x0r

                x0r = x1r + x3i : x0i = x1i - x3r

                a(j3) = wk3r * x0r - wk3i * x0i : a(j3 + 1) = wk3r * x0i + wk3i * x0r
            Next j
        Next k
    End Sub

    Private Sub rftfsub(ByVal a() As Double)
        Dim xr, xi, yr, yi As Double

        Dim ec As Double = 2 * M_PI_2 / n
        Dim wkr As Double = 0.0#
        Dim wki As Double = 0.0#
        Dim wdi As Double = Cos(ec)
        Dim wdr As Double = Sin(ec)

        wdi *= wdr
        wdr *= wdr

        Dim w1r As Double = 1.0# - 2.0# * wdr
        Dim w1i As Double = 2.0# * wdi
        Dim ss As Double = 2.0# * w1i

        Dim i As Integer = n >> 1
        Do
            Dim i0 As Integer = i - 4 * RDFT_LOOP_DIV
            If i0 < 4 Then i0 = 4

            For j As Integer = i - 4 To i0 Step -4
                Dim k As Integer = n - j

                xr = a(j + 2) - a(k - 2) : xi = a(j + 3) + a(k - 1)
                yr = wdr * xr - wdi * xi : yi = wdr * xi + wdi * xr

                a(j + 2) -= yr : a(j + 3) -= yi
                a(k - 2) += yr : a(k - 1) -= yi

                wkr += ss * wdi : wki += ss * (0.5 - wdr)

                xr = a(j) - a(k) : xi = a(j + 1) + a(k + 1)
                yr = wkr * xr - wki * xi : yi = wkr * xi + wki * xr

                a(j) -= yr : a(j + 1) -= yi
                a(k) += yr : a(k + 1) -= yi

                wdr += ss * wki : wdi += ss * (0.5 - wkr)
            Next

            If i0 = 4 Then Exit Do

            wkr = 0.5# * Sin(ec * i0) : wki = 0.5# * Cos(ec * i0)
            wdr = 0.5# - (wkr * w1r - wki * w1i) : wdi = wkr * w1i + wki * w1r

            wkr = 0.5# - wkr
            i = i0
        Loop

        xr = a(2) - a(n - 2) : xi = a(3) + a(n - 1)
        yr = wdr * xr - wdi * xi : yi = wdr * xi + wdi * xr

        a(2) -= yr : a(3) -= yi
        a(n - 2) += yr : a(n - 1) -= yi
    End Sub

    Private Sub rftbsub(ByVal a() As Double)
        Dim xr, xi, yr, yi As Double

        Dim ec As Double = 2 * M_PI_2 / n
        Dim wkr As Double = 0.0#
        Dim wki As Double = 0.0#
        Dim wdi As Double = Cos(ec)
        Dim wdr As Double = Sin(ec)

        wdi *= wdr
        wdr *= wdr

        Dim w1r As Double = 1 - 2 * wdr
        Dim w1i As Double = 2 * wdi
        Dim ss As Double = 2 * w1i

        Dim i As Integer = n >> 1
        a(i + 1) = -a(i + 1)

        Do
            Dim i0 As Integer = i - 4 * RDFT_LOOP_DIV
            If i0 < 4 Then i0 = 4
            For j As Integer = i - 4 To i0 Step -4
                Dim k As Integer = n - j

                xr = a(j + 2) - a(k - 2) : xi = a(j + 3) + a(k - 1)
                yr = wdr * xr + wdi * xi : yi = wdr * xi - wdi * xr

                a(j + 2) -= yr : a(j + 3) = yi - a(j + 3)
                a(k - 2) += yr : a(k - 1) = yi - a(k - 1)

                wkr += ss * wdi : wki += ss * (0.5 - wdr)

                xr = a(j) - a(k) : xi = a(j + 1) + a(k + 1)
                yr = wkr * xr + wki * xi : yi = wkr * xi - wki * xr

                a(j) -= yr : a(j + 1) = yi - a(j + 1)
                a(k) += yr : a(k + 1) = yi - a(k + 1)

                wdr += ss * wki : wdi += ss * (0.5 - wkr)
            Next j

            If i0 = 4 Then Exit Do

            wkr = 0.5# * Sin(ec * i0) : wki = 0.5# * Cos(ec * i0)
            wdr = 0.5# - (wkr * w1r - wki * w1i) : wdi = wkr * w1i + wki * w1r

            wkr = 0.5# - wkr
            i = i0
        Loop

        xr = a(2) - a(n - 2) : xi = a(3) + a(n - 1)
        yr = wdr * xr + wdi * xi : yi = wdr * xi - wdi * xr

        a(2) -= yr : a(3) = yi - a(3)
        a(n - 2) += yr : a(n - 1) = yi - a(n - 1)
        a(1) = -a(1)
    End Sub
End Module
